﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace vcsharp_audio_record_demo
{
    enum tState
    {
        UNDEFINED,
        PAUSED,
        STOPPED,
        PLAYING,
        RECORDING,
        RECORDING_DONE,
        LOADING,
        LOADED
    };
    public partial class Form1 : Form
    {
        /// WaveIn procedures.
        /// Retrieves the number of waveform input devices present in the system.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInGetNumDevs();

        /// Prepares a buffer for waveform input.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInPrepareHeader(IntPtr hwi, tWaveHeader pwh, uint cbwh);

        /// Sends an input buffer to the specified waveform-audio input device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInAddBuffer(IntPtr hwi, tWaveHeader pwh, uint cbwh);

        /// Opens a specified waveform input device for recording.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInOpen(ref IntPtr hwi, uint uDeviceID, tWaveFormatEx lpFormat, uint dwCallback, uint dwInstance, uint dwFlags);

        /// Starts input on the specified waveform input device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInStart(IntPtr hwi);

        /// Stops waveform input.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInStop(IntPtr hwi);

        /// Stops input on a specified waveform input device and resets the current position to 0.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInReset(IntPtr hwi);

        /// Un-prepares the waveIn header.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInUnprepareHeader(IntPtr hwi, tWaveHeader lpWaveInHdr, uint uSize);

        /// Closes the specified waveform-audio input device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveInClose(IntPtr hwi);

        /// WaveOut procedures.
        /// Retrieves the number of waveform output devices present in the system.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutGetNumDevs();

        /// Opens a specified waveform output device for playback.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutOpen(ref IntPtr hWaveOut, uint uDeviceID, tWaveFormatEx lpFormat, uint dwCallback, uint dwInstance, uint dwFlags);

        /// Prepares a waveform data block for playback.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutPrepareHeader(IntPtr hWaveOut, tWaveHeader lpWaveOutHdr, uint uSize);

        /// Sends a data block to the specified waveform output device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutWrite(IntPtr hWaveOut, tWaveHeader lpWaveOutHdr, uint uSize);

        /// Pauses playback on a specified waveform output device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutPause(IntPtr hWaveOut);

        /// Restarts a paused waveform output device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutRestart(IntPtr hWaveOut);

        /// Stops playback on a specified waveform output device and resets the current position to 0.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutReset(IntPtr hWaveOut);

        /// Sets the volume of a waveform output device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutSetVolume(IntPtr hWaveOut, uint dwVolume);

        /// Un-prepares the waveOut header.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutUnprepareHeader(IntPtr hWaveOut, tWaveHeader lpWaveOutHdr, uint uSize);

        /// Closes the specified waveform output device.
        [DllImport("coredll.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern uint waveOutClose(IntPtr hWaveOut);

        /// File format class.
        public class tFileFormat
        {
            public uint fileType;
            public uint fileFormat;
            public uint fileChunkSize;
            public uint formatChunkSize;        ///< Including extra bytes after format chunk.
            public uint dataChunkSize;
        }
        /// Wave header class.
        public class tWaveHeader
        {
            public IntPtr lpData;               ///< Pointer to data buffer.
            public uint dwBufferLength;         ///< Length of data buffer.
            public uint dwBytesRecorded;        ///< Number of bytes recorded in the buffer - used for input only.
            public uint dwUser;                 ///< For client's use.
            public uint dwFlags;                ///< Assorted flags.
            public uint dwLoops;                ///< Loop control counter.
            public IntPtr lpNext;               ///< Reserved for driver.
            public uint reserved;               ///< Reserved for driver.
        }
        /// Audio format class.
        public class tWaveFormatEx
        {
            public ushort wFormatTag;           ///< Format type.
            public ushort nChannels;            ///< Number of channels (mono, stereo).
            public uint nSamplesPerSec;         ///< Sample rate.
            public uint nAvgBytesPerSec;        ///< Average bytes per second.
            public ushort nBlockAlign;          ///< Block size of data.
            public ushort wBitsPerSample;       ///< Number of bits per sample.
            public ushort cbSize;               ///< Extra information size in bytes.
        }
        /// Constants declarations.
        private const uint recordLengthInSecond = 60;
        private const uint formatPCM = 1;
        private const uint successNoError = 0;
        private const uint waveErrorBadFormat = 32;
        private const uint whdrPrepared = 2;
        private const uint deviceMapper = 0xFFFFFFFF;
        private const uint fileTypeRIFF = 0x46464952;       ///< "RIFF" Little Endian format.
        private const uint fileFormatWAVE = 0x45564157;     ///< "WAVE" Little Endian format.
        private const uint formatChunkIdFMT = 0x20746d66;   ///< "fmt " Little Endian format.
        private const uint dataChunkIdDATA = 0x61746164;    ///< "data" Little Endian format.
        private const uint max_volume = 0xFFFFFFFF;         ///< Maximum volume.
        /// Variables declarations.
        private GCHandle gcRecordHandle;
        private GCHandle gcPlaybackHandle;
        private IntPtr handleWaveIn = IntPtr.Zero;
        private IntPtr handleWaveOut = IntPtr.Zero;
        private tFileFormat fileId = new tFileFormat();
        private tWaveFormatEx formatChunkWaveIn = new tWaveFormatEx();
        private tWaveFormatEx formatChunkWaveOut = new tWaveFormatEx();
        private tWaveHeader waveInHeader = new tWaveHeader();
        private tWaveHeader waveOutHeader = new tWaveHeader();
        private tState currentRecordState = tState.UNDEFINED;
        private tState currentPlayState = tState.UNDEFINED;
        private FileStream fileStream = null;
        private BinaryReader binaryReader = null;
        private BinaryWriter binaryWriter = null;
        private uint playbackDurationInSeconds = 0;
        private uint recordingDurationInSeconds = 0;
        private byte[] recordBuffer = null;
        private byte[] playbackBuffer = null;

        public Form1()
        {
            uint numberOfInputDevices = 0;
            uint numberOfOutputDevices = 0;
            uint retErrorCode = 0;
            /// Initializes components.
            InitializeComponent();
            /// Checks for number of input devices present in the system for recording.
            if ((numberOfInputDevices = waveInGetNumDevs()) == 0)
            {
                statusBar.Text = "No input device for recording.";
                return;
            }
            else
            {
                /// Designs format of waveform.
                formatChunkWaveIn.wBitsPerSample = 16;
                formatChunkWaveIn.nSamplesPerSec = 44100;
                formatChunkWaveIn.nChannels = 2;
                formatChunkWaveIn.wFormatTag = 1;
                formatChunkWaveIn.nBlockAlign = (ushort)(formatChunkWaveIn.wBitsPerSample * formatChunkWaveIn.nChannels / 8);
                formatChunkWaveIn.nAvgBytesPerSec = formatChunkWaveIn.nSamplesPerSec * formatChunkWaveIn.nBlockAlign;
                /// Opens a waveform input device for recording purpose.
                retErrorCode = waveInOpen(ref handleWaveIn, deviceMapper, formatChunkWaveIn, 0, 0, 0);
                if (retErrorCode != successNoError)
                {
                    statusBar.Text = "Unable to open the recording device.";
                    return;
                }
                else
                {
                    recordButton.Enabled = true;
                }
            }
            /// Checks for number of output devices present in the system for playback.
            if ((numberOfOutputDevices = waveOutGetNumDevs()) == 0)
            {
                statusBar.Text = "No output device for playback.";
                return;
            }
            else
            {
                openButton.Enabled = true;
            }
        }

        /// Allocates memory (Tries up to minimum of 1/4th of memory to allocate in case of low memory available).
        private bool AllocateMemory(ref byte[] buffer, ref uint allocationSize)
        {
            uint totalBytesToAllocate = 0;
            uint loopCount = 0;

            buffer = null;
            totalBytesToAllocate = allocationSize;
            allocationSize = 0;
            for (loopCount = 0; loopCount <= 2; loopCount++)
            {
                try
                {
                    buffer = new byte[totalBytesToAllocate];
                }
                catch (OutOfMemoryException)
                {
                    totalBytesToAllocate /= 2;
                    continue;
                }
                break;
            }
            if (buffer == null)
            {
                return false;
            }
            allocationSize = totalBytesToAllocate;
            return true;
        }

        /// Enable/Disables the buttons based on current state.
        private void SetButtonsState(tState state)
        {
            switch (state)
            {
                case tState.RECORDING:
                    {
                        recordButton.Enabled = false;
                        stopRecordButton.Enabled = true;
                        saveButton.Enabled = false;
                        break;
                    }
                case tState.RECORDING_DONE:
                    {
                        recordButton.Enabled = true;
                        stopRecordButton.Enabled = false;
                        saveButton.Enabled = true;
                        break;
                    }
                case tState.LOADING:
                    {
                        pauseButton.Enabled = false;
                        stopButton.Enabled = false;
                        break;
                    }
                case tState.LOADED:
                    {
                        playButton.Enabled = true;
                        pauseButton.Enabled = false;
                        stopButton.Enabled = false;
                        break;
                    }
                case tState.PLAYING:
                    {
                        playButton.Enabled = false;
                        pauseButton.Enabled = true;
                        stopButton.Enabled = true;
                        break;
                    }
                case tState.PAUSED:
                    {
                        playButton.Enabled = true;
                        pauseButton.Enabled = false;
                        stopButton.Enabled = false;
                        break;
                    }
                case tState.STOPPED:
                    {
                        playButton.Enabled = true;
                        pauseButton.Enabled = false;
                        stopButton.Enabled = false;
                        break;
                    }
                default:
                    {
                        playButton.Enabled = false;
                        pauseButton.Enabled = false;
                        stopButton.Enabled = false;
                        break;
                    }
            }
        }

        /// Opens a file and prepares format and data buffer for playback.
        private void openButton_Click(object sender, EventArgs e)
        {
            uint retErrorCode = successNoError;
            byte[] formatId = null;
            uint tempId = 0;
            bool loopTerminate = false;

            /// File formats to filter from the open-select window.
            openFileDialog.Filter = "Wave Files (.WAV)|*.WAV";
            openFileDialog.FileName = ".wav";
            DialogResult result = openFileDialog.ShowDialog();
            if (result == DialogResult.OK)
            {
                playbackTimer.Enabled = false;
                statusBar.Text = "File loading... (Please wait)";
                /// Enables/Disables the buttons.
                SetButtonsState(tState.UNDEFINED);
                /// Stops the playback, if running.
                if (currentPlayState == tState.PLAYING || currentPlayState == tState.PAUSED)
                {
                    waveOutReset(handleWaveOut);
                    waveOutHeader.dwFlags &= 0x0E;
                    currentPlayState = tState.STOPPED;
                }
                /// Un-pins the byte array, if pinned.
                if (gcPlaybackHandle.IsAllocated == true)
                {
                    waveOutHeader.lpData = IntPtr.Zero;
                    gcPlaybackHandle.Free();
                }
                /// Un-prepares the header, if prepared for output.
                if (((waveOutHeader.dwFlags & whdrPrepared) != 0) || (waveOutHeader.reserved != 0))
                {
                    waveOutUnprepareHeader(handleWaveOut, waveOutHeader, (uint)Marshal.SizeOf(waveOutHeader));
                }
                /// Closes the waveform-audio output device, if open.
                if (handleWaveOut != IntPtr.Zero)
                {
                    waveOutClose(handleWaveOut);
                }
                /// Clear waveOutHeader.
                waveOutHeader.dwBufferLength = 0;
                waveOutHeader.dwBytesRecorded = 0;
                waveOutHeader.dwFlags = 0x00;
                /// Opens selected file for playback.
                try
                {
                    fileStream = new FileStream(openFileDialog.FileName, FileMode.Open, FileAccess.Read, FileShare.Read);
                }
                catch (IOException)
                {
                    statusBar.Text = "Unable to open - Selected file in use.";
                    return;
                }
                catch (UnauthorizedAccessException)
                {
                    statusBar.Text = "Unable to open - Unauthorized access denied.";
                    return;
                }
                catch
                {
                    statusBar.Text = "Unable to open the file.";
                    return;
                }
                binaryReader = new BinaryReader(fileStream);
                /// Reads file type and format from file header.
                fileId.fileType = binaryReader.ReadUInt32();
                fileId.fileChunkSize = binaryReader.ReadUInt32();
                fileId.fileFormat = binaryReader.ReadUInt32();
                /// If format chunk id not found.
                if (binaryReader.ReadUInt32() != formatChunkIdFMT)
                {
                    statusBar.Text = "File format not supported.";
                    return;
                }
                /// Read format Chunk.
                fileId.formatChunkSize = binaryReader.ReadUInt32();
                formatChunkWaveOut.wFormatTag = binaryReader.ReadUInt16();
                formatChunkWaveOut.nChannels = binaryReader.ReadUInt16();
                formatChunkWaveOut.nSamplesPerSec = binaryReader.ReadUInt32();
                formatChunkWaveOut.nAvgBytesPerSec = binaryReader.ReadUInt32();
                formatChunkWaveOut.nBlockAlign = binaryReader.ReadUInt16();
                formatChunkWaveOut.wBitsPerSample = binaryReader.ReadUInt16();
                /// Show format of selected file.
                formatBar.Text = "--> Playback Rate";
                formatId = BitConverter.GetBytes(fileId.fileType);
                formatBar.Text += "\r\nFile type: " + System.Text.Encoding.UTF8.GetString(formatId, 0, 4);
                formatId = BitConverter.GetBytes(fileId.fileFormat);
                formatBar.Text += "\r\nFile format: " + System.Text.Encoding.UTF8.GetString(formatId, 0, 4);
                formatBar.Text += "\r\nAudio format: " + formatChunkWaveOut.wFormatTag.ToString();
                formatBar.Text += "\r\nTotal channels: " + formatChunkWaveOut.nChannels.ToString();
                formatBar.Text += "\r\nBlock alignment: " + formatChunkWaveOut.nBlockAlign.ToString();
                formatBar.Text += "\r\nBits per sample: " + formatChunkWaveOut.wBitsPerSample.ToString();
                formatBar.Text += "\r\nSamples per second: " + formatChunkWaveOut.nSamplesPerSec.ToString();
                formatBar.Text += "\r\nAvarage bytes per second: " + formatChunkWaveOut.nAvgBytesPerSec.ToString();
                /// Jumps to field next to format chunk.
                fileStream.Seek(fileId.formatChunkSize - 16, SeekOrigin.Current);   ///< Format chunk size for PCM = 16.
                /// Search for "data" id in the file.
                while (loopTerminate == false)
                {
                    try
                    {
                        tempId = binaryReader.ReadUInt32();
                    }
                    catch
                    {
                        loopTerminate = true;
                        continue;
                    }
                    if (tempId == dataChunkIdDATA)
                    {
                        break;
                    }
                    else
                    {
                        try
                        {
                            fileStream.Seek(binaryReader.ReadUInt32(), SeekOrigin.Current);
                        }
                        catch
                        {
                            loopTerminate = true;
                            continue;
                        }
                    }
                }
                if (loopTerminate == true)
                {
                    statusBar.Text = "File corrupted.";
                    return;
                }
                fileId.dataChunkSize = binaryReader.ReadUInt32();
                try
                {
                    playbackBuffer = null;
                    playbackBuffer = new byte[fileId.dataChunkSize];
                }
                catch (OutOfMemoryException)
                {
                    statusBar.Text = "File reading failed - Memory not enough.";
                    return;
                }
                catch
                {
                    statusBar.Text = "Unable to read the file.";
                    return;
                }
                /// Pin the byte array to avoid Garbage collector moving it.
                gcPlaybackHandle = GCHandle.Alloc(playbackBuffer, GCHandleType.Pinned);
                waveOutHeader.lpData = gcPlaybackHandle.AddrOfPinnedObject();
                fileStream.Read(playbackBuffer, 0, (int)fileId.dataChunkSize);
                waveOutHeader.dwBufferLength = fileId.dataChunkSize;
                /// Closes the open file stream.
                binaryReader.Close();
                fileStream.Close();
                statusBar.Text = "File loaded - " + Path.GetFileName(openFileDialog.FileName);
                /// Opens a waveform output device for playback.
                retErrorCode = waveOutOpen(ref handleWaveOut, deviceMapper, formatChunkWaveOut, 0, 0, 0);
                if (retErrorCode != successNoError)
                {
                    if (retErrorCode == waveErrorBadFormat)
                    {
                        statusBar.Text = "File Format not supported.";
                    }
                    else
                    {
                        statusBar.Text = "Unable to open the device for playback.";
                    }
                    return;
                }
                retErrorCode = waveOutPrepareHeader(handleWaveOut, waveOutHeader, (uint)Marshal.SizeOf(waveOutHeader));
                if (retErrorCode != successNoError)
                {
                    statusBar.Text = "Unable to open the device for playback.";
                    return;
                }
                currentPlayState = tState.LOADED;
                /// Enable/Disables the buttons.
                SetButtonsState(currentPlayState);
            }
        }

        private void playButton_Click(object sender, EventArgs e)
        {
            uint retErrorCode = successNoError;
            if (currentPlayState == tState.PAUSED)
            {
                waveOutRestart(handleWaveOut);
                statusBar.Text = "Resuming...";
            }
            else
            {
                waveOutHeader.dwFlags &= 0x0E;
                /// Sends a data block to the specified waveform output device.
                retErrorCode = waveOutWrite(handleWaveOut, waveOutHeader, (uint)Marshal.SizeOf(waveOutHeader));
                if (retErrorCode != successNoError)
                {
                    statusBar.Text = "Unable to play the file.";
                    return;
                }
                waveOutSetVolume(handleWaveOut, max_volume);
                statusBar.Text = "Playing...";
                playbackDurationInSeconds = fileId.dataChunkSize / formatChunkWaveOut.nAvgBytesPerSec;
            }
            playbackTimer.Enabled = true;
            currentPlayState = tState.PLAYING;
            /// Enable/Disables the buttons.
            SetButtonsState(currentPlayState);
        }

        private void pauseButton_Click(object sender, EventArgs e)
        {
            playbackTimer.Enabled = false;
            /// Pauses playback on a specified waveform output device.
            waveOutPause(handleWaveOut);
            currentPlayState = tState.PAUSED;
            /// Enables/Disables the buttons.
            SetButtonsState(currentPlayState);
            statusBar.Text = "Paused.";
        }

        private void stopButton_Click(object sender, EventArgs e)
        {
            if (playbackTimer.Enabled == false)
            {
                statusBar.Text = "Playback completed.";
            }
            else
            {
                playbackTimer.Enabled = false;
                statusBar.Text = "Stopped.";
            }
            /// Stops playback on a specified waveform output device.
            waveOutReset(handleWaveOut);
            /// Clears WHDR_DONE and WHDR_INQUEUE flags.
            waveOutHeader.dwFlags &= 0x0E;
            currentPlayState = tState.STOPPED;
            /// Enable/Disables the buttons.
            SetButtonsState(currentPlayState);
        }

        private void recordButton_Click(object sender, EventArgs e)
        {
            uint totalBytesToAllocate = 0;
            bool retVal = false;
            uint retErrorCode = successNoError;

            totalBytesToAllocate = formatChunkWaveIn.nAvgBytesPerSec * recordLengthInSecond;
            /// Un-prepares the header, if prepared for input.
            if (((waveInHeader.dwFlags & whdrPrepared) != 0) || (waveInHeader.reserved != 0))
            {
                retErrorCode = waveInUnprepareHeader(handleWaveIn, waveInHeader, (uint)Marshal.SizeOf(waveInHeader));
            }
            /// Un-pins the byte array, if pinned.
            if (gcRecordHandle.IsAllocated == true)
            {
                waveInHeader.lpData = IntPtr.Zero;
                gcRecordHandle.Free();
            }
            /// Clears waveInHeader.
            waveInHeader.dwFlags = 0x00;
            waveInHeader.dwBytesRecorded = 0;
            waveInHeader.dwBufferLength = 0;
            /// Allocates memory for recording.
            retVal = AllocateMemory(ref recordBuffer, ref totalBytesToAllocate);
            if (retVal == false)
            {
                statusBar.Text = "Memory not enough for recording.";
                return;
            }
            /// Pin the byte array to avoid Garbage collector moving it .
            gcRecordHandle = GCHandle.Alloc(recordBuffer, GCHandleType.Pinned);
            waveInHeader.lpData = gcRecordHandle.AddrOfPinnedObject();
            waveInHeader.dwBufferLength = totalBytesToAllocate;
            recordingDurationInSeconds = totalBytesToAllocate / formatChunkWaveIn.nAvgBytesPerSec;
            /// Prepares a buffer for waveform input.
            retErrorCode = waveInPrepareHeader(handleWaveIn, waveInHeader, (uint)Marshal.SizeOf(waveInHeader));
            if (retErrorCode != successNoError)
            {
                statusBar.Text = "Unable to start the recording.";
                return;
            }
            /// Sends buffer to the waveform input device.
            retErrorCode = waveInAddBuffer(handleWaveIn, waveInHeader, (uint)Marshal.SizeOf(waveInHeader));
            if (retErrorCode != successNoError)
            {
                statusBar.Text = "Unable to start the recording.";
                return;
            }
            /// Starts input on the waveform input device.
            retErrorCode = waveInStart(handleWaveIn);
            if (retErrorCode != successNoError)
            {
                statusBar.Text = "Unable to start the recording.";
                return;
            }
            recordingTimer.Enabled = true;
            /// Show format of selected file.
            formatBar.Text = "--> Recording Rate";
            formatBar.Text += "\r\nFile type: RIFF";
            formatBar.Text += "\r\nFile format: WAVE";
            formatBar.Text += "\r\nAudio format: " + formatChunkWaveIn.wFormatTag.ToString();
            formatBar.Text += "\r\nTotal channels: " + formatChunkWaveIn.nChannels.ToString();
            formatBar.Text += "\r\nBlock alignment: " + formatChunkWaveIn.nBlockAlign.ToString();
            formatBar.Text += "\r\nBits per sample: " + formatChunkWaveIn.wBitsPerSample.ToString();
            formatBar.Text += "\r\nSamples per second: " + formatChunkWaveIn.nSamplesPerSec.ToString();
            formatBar.Text += "\r\nAvarage bytes per second: " + formatChunkWaveIn.nAvgBytesPerSec.ToString();
            statusBar.Text = "Recording... (Total duration = " + (totalBytesToAllocate / formatChunkWaveIn.nAvgBytesPerSec).ToString() + " Seconds)";
            currentRecordState = tState.RECORDING;
            /// Enable/Disables the buttons.
            SetButtonsState(currentRecordState);
        }

        private void stopRecordBotton_Click(object sender, EventArgs e)
        {
            if (recordingTimer.Enabled == false)
            {
                statusBar.Text = "Recording done (Timeout).";
            }
            else
            {
                recordingTimer.Enabled = false;
                statusBar.Text = statusBar.Text = "Recording done.";
            }
            /// Stops input on waveform input device.
            waveInStop(handleWaveIn);
            /// Stops input on waveform input device and resets the current position to 0.
            waveInReset(handleWaveIn);
            currentRecordState = tState.RECORDING_DONE;
            /// Enable/Disables the buttons.
            SetButtonsState(currentRecordState);
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            string newExtension = string.Empty;
            /// File formats to filter from the save-as window.
            saveFileDialog.Filter = "Wave Files (.WAV)|*.wav";
            saveFileDialog.FileName = ".wav";
            DialogResult result = saveFileDialog.ShowDialog();
            if (result == DialogResult.OK)
            {
                /// Change file extension, if not ".wav".
                saveFileDialog.FileName = Path.ChangeExtension(saveFileDialog.FileName, ".wav");
                try
                {
                    fileStream = new FileStream(saveFileDialog.FileName, FileMode.Create, FileAccess.Write, FileShare.None);
                }
                catch (IOException)
                {
                    statusBar.Text = "Unable to save - Selected file in use.";
                    return;
                }
                catch (UnauthorizedAccessException)
                {
                    statusBar.Text = "Unable to save - Access denied.";
                    return;
                }
                catch
                {
                    statusBar.Text = "Unable to save - Unable to open the file.";
                    return;
                }
                statusBar.Text = "File saving... (Please wait)";
                binaryWriter = new BinaryWriter(fileStream);
                fileId.fileChunkSize = (uint)Marshal.SizeOf(fileFormatWAVE) +
                             (uint)Marshal.SizeOf(formatChunkIdFMT) +
                             (uint)Marshal.SizeOf(fileId.formatChunkSize) + 16 /*Format chunk size for PCM = 16*/ +
                             (uint)Marshal.SizeOf(dataChunkIdDATA) +
                             (uint)Marshal.SizeOf(fileId.dataChunkSize) +
                             waveInHeader.dwBytesRecorded;
                fileId.dataChunkSize = waveInHeader.dwBytesRecorded;
                binaryWriter.Write(fileTypeRIFF);
                binaryWriter.Write(fileId.fileChunkSize);
                binaryWriter.Write(fileFormatWAVE);
                binaryWriter.Write(formatChunkIdFMT);
                binaryWriter.Write((uint)16);         ///< Format chunk size for PCM = 16.
                binaryWriter.Write(formatChunkWaveIn.wFormatTag);
                binaryWriter.Write(formatChunkWaveIn.nChannels);
                binaryWriter.Write(formatChunkWaveIn.nSamplesPerSec);
                binaryWriter.Write(formatChunkWaveIn.nAvgBytesPerSec);
                binaryWriter.Write(formatChunkWaveIn.nBlockAlign);
                binaryWriter.Write(formatChunkWaveIn.wBitsPerSample);
                binaryWriter.Write(dataChunkIdDATA);
                binaryWriter.Write(fileId.dataChunkSize);
                binaryWriter.Write(recordBuffer, 0, (int)waveInHeader.dwBytesRecorded);
                /// Closes the open file stream.
                binaryWriter.Close();
                fileStream.Close();
                statusBar.Text = "File saved - " + Path.GetFileName(saveFileDialog.FileName);
            }
        }

        private void Form1_closing(object sender, EventArgs e)
        {
            if (currentRecordState == tState.RECORDING)
            {
                /// Stops input on waveform input device.
                waveInStop(handleWaveIn);
                /// Stops input on waveform input device and resets the current position to 0.
                waveInReset(handleWaveIn);
            }
            if (currentPlayState == tState.PLAYING || currentPlayState == tState.PAUSED)
            {
                /// Stops playback on a specified waveform output device.
                waveOutReset(handleWaveOut);
            }
            /// Un-prepares the header, if prepared for input.
            if (((waveInHeader.dwFlags & whdrPrepared) != 0) || (waveInHeader.reserved != 0))
            {
                waveInUnprepareHeader(handleWaveIn, waveInHeader, (uint)Marshal.SizeOf(waveInHeader));
            }
            /// Un-prepares the header, if prepared for output.
            if (((waveOutHeader.dwFlags & whdrPrepared) != 0) || (waveOutHeader.reserved != 0))
            {
                waveOutUnprepareHeader(handleWaveOut, waveOutHeader, (uint)Marshal.SizeOf(waveOutHeader));
            }
            /// Unpin the array, if pinned.
            if (gcRecordHandle.IsAllocated == true)
            {
                waveInHeader.lpData = IntPtr.Zero;
                gcRecordHandle.Free();
            }
            recordBuffer = null;
            /// Unpin the array, if pinned.
            if (gcPlaybackHandle.IsAllocated == true)
            {
                waveOutHeader.lpData = IntPtr.Zero;
                gcPlaybackHandle.Free();
            }
            playbackBuffer = null;
            /// Closes the waveform-audio input device, if open.
            if (handleWaveIn != IntPtr.Zero)
            {
                waveInClose(handleWaveIn);
            }
            /// Closes the waveform-audio output device, if open.
            if (handleWaveOut != IntPtr.Zero)
            {
                waveOutClose(handleWaveOut);
            }
        }
        /// Timer to reset the playback when finished.
        private void playbackTimer_Tick(object sender, EventArgs e)
        {
            if (playbackDurationInSeconds > 0)
            {
                playbackDurationInSeconds--;
            }
            else
            {
                playbackTimer.Enabled = false;
                stopButton_Click(sender, e);
            }
        }
        /// Timer to stop the recording after time-out.
        private void recordingTimer_Tick(object sender, EventArgs e)
        {
            if (recordingDurationInSeconds > 0)
            {
                recordingDurationInSeconds--;
            }
            else
            {
                recordingTimer.Enabled = false;
                stopRecordBotton_Click(sender, e);
            }
        }
    }
}